package com.sagacloud.util.compute_engine.core;

import com.sagacloud.parser.simulator.String2Grammar;
import com.sagacloud.util.compute_engine.util.DateUtil;

import java.text.SimpleDateFormat;
import java.util.*;

public class CoreComputeUtilUtil {
    public static void main(String[] args) throws Exception {
        String expression = "servicedata_1001_10101+servicedata_1002_10101-servicedata_1003_10101-servicedata_1004_10101";
        GrammarTreeNode node = String2Grammar.Generate(expression);
        Map<String, Double> dataMap = new HashMap<String, Double>();
        dataMap.put("servicedata_1001_10101", null);
        dataMap.put("servicedata_1002_10101", 0.0);
        dataMap.put("servicedata_1003_10101", 0.0);
        dataMap.put("servicedata_1003_10101", null);
        boolean rrr = CantCompute(node, dataMap);
        System.out.println(rrr);
    }

    public static void BuildNetwork(CoreInstance instance) throws Exception {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss");
        List<CoreOutput> outputListOrdered = new ArrayList<CoreOutput>();
        List<CoreInput> inputList = instance.inputList;
        List<CoreOutput> outputList = instance.outputList;
        Map<String, CoreInput> inputMap = new HashMap<String, CoreInput>();
        for (CoreInput input : inputList) {
            input.successorList = new ArrayList<CoreOutput>();
            inputMap.put(input.input_sign, input);
        }
        Map<String, CoreOutput> outputMap = new HashMap<String, CoreOutput>();
        Map<String, List<CoreOutput>> sign2listMap = new HashMap<String, List<CoreOutput>>();
        for (CoreOutput output : outputList) {
            output.predecessorList = new ArrayList<Object>();
            output.successorList = new ArrayList<CoreOutput>();
            output.all_input = new HashMap<String, Boolean>();
            outputMap.put(output.output_sign + "-" + sdf.format(output.time_begin), output);
            if (!sign2listMap.containsKey(output.output_sign)) {
                sign2listMap.put(output.output_sign, new ArrayList<CoreOutput>());
            }
            sign2listMap.get(output.output_sign).add(output);
        }
        for (CoreOutput output : outputList) {
            List<CoreInput> tmpInputList = new ArrayList<CoreInput>();
            List<CoreOutput> tmpOutputList = new ArrayList<CoreOutput>();
            for (String input_sign : output.var_signMap.keySet()) {
                if (inputMap.containsKey(input_sign)) {
                    CoreInput input = inputMap.get(input_sign);
                    tmpInputList.add(input);
                } else {
                    CoreOutput tmp_output = outputMap.get(input_sign + "-" + sdf.format(output.time_begin));
                    if (tmp_output == null) {
                        List<CoreOutput> relList = sign2listMap.get(input_sign);
                        if (relList.size() == 1) {
                            tmp_output = relList.get(0);
                        }
                    }
                    tmpOutputList.add(tmp_output);
                }
            }
            for (CoreInput tmpInput : tmpInputList) {
                tmpInput.successorList.add(output);
                output.predecessorList.add(tmpInput);
            }
            for (CoreOutput tmpOutput : tmpOutputList) {
                tmpOutput.successorList.add(output);
                output.predecessorList.add(tmpOutput);
            }
        }
        Map<String, Boolean> finishMap = new HashMap<String, Boolean>();
        while (true) {
            int size1 = finishMap.size();
            for (CoreOutput output : outputList) {
                if (finishMap.containsKey(output.output_sign + "_" + sdf.format(output.time_begin))) {
                    continue;
                }
                List<CoreOutput> tmpOutputList = new ArrayList<CoreOutput>();
                for (Object predecessor : output.predecessorList) {
                    if (predecessor instanceof CoreOutput) {
                        tmpOutputList.add((CoreOutput) predecessor);
                    }
                }
                boolean children_ok = true;
                for (CoreOutput tmpOutput : tmpOutputList) {
                    if (!finishMap.containsKey(tmpOutput.output_sign + "_" + sdf.format(tmpOutput.time_begin))) {
                        children_ok = false;
                        break;
                    }
                }
                if (!children_ok) {
                    continue;
                }
                for (Object predecessor : output.predecessorList) {
                    if (predecessor instanceof CoreInput) {
                        CoreInput input = (CoreInput) predecessor;
                        output.all_input.putIfAbsent(input.input_sign, true);
                        if (!output.time_period_intialized) {
                            if (DateUtil.Instance().compare(output.time_period, input.time_period) < 0) {
                                output.time_period = input.time_period;
                            }
                        }
                    } else if (predecessor instanceof CoreOutput) {
                        CoreOutput tmpOutput = (CoreOutput) predecessor;
                        if (!output.time_period_intialized) {
                            if (DateUtil.Instance().compare(output.time_period, tmpOutput.time_period) < 0) {
                                output.time_period = tmpOutput.time_period;
                            }
                        }
                    }
                }
                for (CoreOutput tmpOutput : tmpOutputList) {
                    for (String tmpkey : tmpOutput.all_input.keySet()) {
                        output.all_input.putIfAbsent(tmpkey, true);
                    }
                }
                finishMap.put(output.output_sign + "_" + sdf.format(output.time_begin), true);
                outputListOrdered.add(output);
            }
            int size2 = finishMap.size();
            if (size2 == size1) {
                break;
            }
        }
        if (finishMap.size() != outputList.size()) {
            StringBuffer sb = new StringBuffer();
            for (CoreOutput output : outputList) {
                if (!finishMap.containsKey(output.output_sign + "_" + sdf.format(output.time_begin))) {
                    sb.append("\t" + output.output_sign + "_" + sdf.format(output.time_begin));
                }
            }
            throw new Exception("CoreInstance network cant build:" + sb);
        }
        instance.outputList = outputListOrdered;
    }

    public static boolean CantCompute(GrammarTreeNode node, Map<String, Double> dataMap) {
        Map<String, Boolean> jiashuMap = new HashMap<String, Boolean>();
        boolean jiahao = jiahao(node, jiashuMap);
        if (jiahao) {
            boolean exist_data = false;
            Iterator<String> iter = jiashuMap.keySet().iterator();
            while (iter.hasNext()) {
                String key = iter.next();
                if (dataMap.get(key) != null) {
                    exist_data = true;
                    break;
                }
            }
            return !exist_data;
        } else {
            return false;
        }
    }

    private static boolean jiahao(GrammarTreeNode node, Map<String, Boolean> jiashuMap) {
        if ("evaluate".equals(node.type)) {
            if ("+".equals(node.evaluate)) {
                boolean all_true = true;
                for (int i = 0; i < node.evaluateItemList.size(); i++) {
                    GrammarTreeNode jia_node = node.evaluateItemList.get(i);
                    if ("evaluate".equals(jia_node.type)) {
                        boolean tmp = jiahao(jia_node, jiashuMap);
                        if (!tmp) {
                            all_true = false;
                        }
                    } else if ("const".equals(jia_node.type)) {
                    } else if ("var".equals(jia_node.type)) {
                        jiashuMap.put(jia_node.var_original, true);
                    } else {
                        all_true = false;
                    }
                }
                return all_true;
            } else if ("-".equals(node.evaluate)) {
                GrammarTreeNode jia_node = node.evaluateItemList.get(0);
                if ("evaluate".equals(jia_node.type)) {
                    return jiahao(jia_node, jiashuMap);
                } else if ("const".equals(jia_node.type)) {
                    return true;
                } else if ("var".equals(jia_node.type)) {
                    jiashuMap.put(jia_node.var_original, true);
                    return true;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}
