package com.sagacloud.util.compute_engine.fjd;

import cn.hutool.core.thread.ThreadUtil;
import com.sagacloud.repository.HBaseSaveManagerRepository;
import com.sagacloud.util.compute_engine.core.Dao_dataManager;
import com.sagacloud.util.compute_engine.core.Dao_data_maxManager;
import com.sagacloud.util.compute_engine.core.IDao_log;
import com.sagacloud.util.compute_engine.util.ConstantCE;
import com.sagacloud.util.compute_engine.util.QueryUtil;
import com.zillion.database.agent.ZillionAgent;
import lombok.extern.slf4j.Slf4j;
import org.apache.hadoop.conf.Configuration;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

@Slf4j
public class ComputeEngineMain {
    List<ThreadCompute> threadList;
    List<ThreadComputeJob> threadJobList;
    ThreadDelay threadDelay;

    public void initDatabaseAndTable() throws Exception {
        if (!QueryUtil.database_exist(ConstantCE.database_name)) {
            HBaseSaveManagerRepository.randomZillionAgent().NDL_AddDatabase(ConstantCE.database_name);
        }
        TableUtil.AddTables(ConstantCE.database_name);
        ThreadUtil.safeSleep(10 * 1000);
        TableUtil.AlterTables(ConstantCE.database_name);

        //可能是分局反射扩展jar包的任务设计，作废注释掉
        Dao_dataManager.Instance().Initialize(ConstantCE.database_name);
        Dao_data_maxManager.Instance().Initialize(ConstantCE.database_name);

        IDao_log dao_log = new ImplDao_log(ConstantCE.database_name);
        threadList = new ArrayList<ThreadCompute>();
        for (int i = 0; i < ConstantCE.thread_count; i++) {
            ThreadCompute thread = new ThreadCompute(ConstantCE.database_name, dao_log);
            thread.start();
            threadList.add(thread);
        }

        if (ConstantCE.job_split) {
            threadJobList = new ArrayList<ThreadComputeJob>();
            for (int i = 0; i < ConstantCE.job_thread_count; i++) {
                ThreadComputeJob thread = new ThreadComputeJob();
                thread.start();
                threadJobList.add(thread);
            }
        }

        if (ConstantCE.original_delay_enable) {
            TableUtil.AddTable(ConstantCE.database_name, "original_delaylog");
            TableUtil.AddTable(ConstantCE.database_name, "original_delaydetail");
            TableUtil.AddTable(ConstantCE.database_name, "original_delaystat");

            threadDelay = new ThreadDelay(ConstantCE.database_name);
            threadDelay.start();
        }
    }

    public ZillionAgent initZillionAgent() throws Exception {
        Configuration configuration = new Configuration();
        InputStream inCoreSite = ConstantCE.getStream("core-site.xml");
        configuration.addResource(inCoreSite);
        InputStream inHbaseSite = ConstantCE.getStream("hbase-site.xml");
        configuration.addResource(inHbaseSite);
        InputStream inHdfsSite = ConstantCE.getStream("hdfs-site.xml");
        configuration.addResource(inHdfsSite);
        ZillionAgent agent = new ZillionAgent(configuration);
        agent.Start();
        return agent;
    }

    public int stop(int exitCode) {
        log.warn("service stopping");

        if (ConstantCE.original_delay_enable) {
            threadDelay.requestStop();
            while (true) {
                try {
                    Thread.sleep(1L);
                } catch (InterruptedException e) {
                    log.error(e.getMessage(),e);
                }
                if (threadDelay.isAlive()) {
                    continue;
                } else {
                    break;
                }
            }
        }

        for (int i = 0; i < threadList.size(); i++) {
            ThreadCompute thread = threadList.get(i);
            thread.requestStop();
        }
        while (true) {
            try {
                Thread.sleep(1L);
            } catch (InterruptedException e) {
                log.error(e.getMessage(),e);
            }

            boolean all_finish = true;
            for (ThreadCompute thread : threadList) {
                if (thread.isAlive()) {
                    all_finish = false;
                    break;
                }
            }
            if (all_finish) {
                break;
            }
        }


        log.warn("service stopped");

        return exitCode;
    }


}
