package com.sagacloud.util.split;

import com.sagacloud.util.compute_engine.util.ConstantCE;
import com.sagacloud.util.compute_engine.util.DoubleTimeObject;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

//ArrangeGeneralManager的辅助类
public class ArrangeUtil {
    private static final ArrangeUtil instance = new ArrangeUtil();

    private ArrangeUtil() {

    }

    public static ArrangeUtil Instance() {
        return instance;
    }

    public void Filter_monotony(List<DoubleTimeObject> recordDataList, List<DoubleTimeObject> badDataList) {
        // 第一次扫描，将绝对值小于前一个的index放到badIndexList
        List<Integer> badIndexList = new ArrayList<Integer>();
        for (int index = 1; index < recordDataList.size(); index++) {
            DoubleTimeObject recordData = recordDataList.get(index);
            DoubleTimeObject recordDataLast = recordDataList.get(index - 1);
            if (Math.abs(recordData.data_value) < Math.abs(recordDataLast.data_value)) {
                badIndexList.add(index);
            }
        }

        for (int i = badIndexList.size() - 1; i >= 0; i--) {
            int badIndex = badIndexList.get(i);
            if (badIndex >= recordDataList.size()) {
                continue;
            }

            // 寻找清理方案
            int[][] offsetArrayArray = {{1, 0}, {0, 1}, {1, 1}, {2, 0}, {0, 2}};
            boolean match_find = false;
            for (int ii = 0; ii < offsetArrayArray.length; ii++) {
                int[] offsetArray = offsetArrayArray[ii];
                int left_start = badIndex - 1 - offsetArray[0];
                int right_start = badIndex + offsetArray[1];
                int[] ttt = {left_start - 3, left_start - 2, left_start - 1, left_start, right_start, right_start + 1, right_start + 2,
                        right_start + 3};
                boolean match = true;
                int cmp_count = 0;
                for (int iii = 1; iii < ttt.length; iii++) {
                    if (ttt[iii - 1] >= 0 && ttt[iii] < recordDataList.size()) {
                        DoubleTimeObject recordData = recordDataList.get(ttt[iii]);
                        DoubleTimeObject recordDataLast = recordDataList.get(ttt[iii - 1]);
                        if (Math.abs(recordData.data_value) < Math.abs(recordDataLast.data_value)) {
                            match = false;
                            break;
                        } else {
                            cmp_count++;
                        }
                    }
                }
                if (match && cmp_count > 2) {
                    for (int iii = right_start - 1; iii >= left_start + 1; iii--) {
                        DoubleTimeObject data = recordDataList.remove(iii);
                        badDataList.add(data);
                    }
                    match_find = true;
                    break;
                }
            }
            if (!match_find) {
            }
        }
    }

    public void FilterMaxMin(long FunctionId, List<DoubleTimeObject> recordDataList, List<DoubleTimeObject> badDataList) {
        if (!ConstantCE.OriginalDataFilterMaxMap.containsKey(FunctionId) && !ConstantCE.OriginalDataFilterMinMap.containsKey(FunctionId)) {
            return;
        }

        List<Integer> badIndexList = new ArrayList<Integer>();
        for (int index = 0; index < recordDataList.size(); index++) {
            DoubleTimeObject recordData = recordDataList.get(index);
            boolean bad = false;
            if (recordData.data_value == null) {
                bad = true;
            } else {
                if (ConstantCE.OriginalDataFilterMaxMap.containsKey(FunctionId)) {
                    double max = ConstantCE.OriginalDataFilterMaxMap.get(FunctionId);
                    if (recordData.data_value > max) {
                        bad = true;
                    }
                }
                if (ConstantCE.OriginalDataFilterMinMap.containsKey(FunctionId)) {
                    double min = ConstantCE.OriginalDataFilterMinMap.get(FunctionId);
                    if (recordData.data_value < min) {
                        bad = true;
                    }
                }
                if (ConstantCE.OriginalDataFilterValuesMap.containsKey(FunctionId)) {
                    List<Double> bad_values = ConstantCE.OriginalDataFilterValuesMap.get(FunctionId);
                    for (Double value : bad_values) {
                        if (value.equals(recordData.data_value)) {
                            bad = true;
                            break;
                        }
                    }
                }
            }
            if (bad) {
                badIndexList.add(index);
            }
        }

        // 从后往前，将标记错数从recordDataList中移除
        for (int index = badIndexList.size() - 1; index >= 0; index--) {
            int badIndex = badIndexList.get(index);
            DoubleTimeObject badData = recordDataList.remove(badIndex);
            badDataList.add(badData);
        }
    }

    public List<DoubleTimeObject> ComputeInterval(long funcid, String data_type, List<DoubleTimeObject> recordDataList,
                                                  List<DoubleTimeObject> badDataList, Date timeFrom, Date timeTo, long millisecondsInterval) throws Exception {
        boolean fjd_interpolate = true;
        if ("Acc".equals(data_type)) {
            fjd_interpolate = true;
        } else if ("Inst".equals(data_type)) {
            fjd_interpolate = ConstantCE.Inst_fjd_interpolate;
        }
        for (int index_remove = recordDataList.size() - 1; index_remove >= 0; index_remove--) {
            DoubleTimeObject recordData = recordDataList.get(index_remove);
            if (recordData.data_value == null) {
                badDataList.add(recordData);
                recordDataList.remove(index_remove);
            }
        }
        this.FilterMaxMin(funcid, recordDataList, badDataList);
        if ("Acc".equals(data_type)) {
            this.Filter_monotony(recordDataList, badDataList);
        }

        List<DoubleTimeObject> result = new ArrayList<DoubleTimeObject>();

        int tmpIndex = 0;
        for (Date tmpDate = timeFrom; tmpDate.compareTo(timeTo) <= 0; tmpDate = new Date(tmpDate.getTime() + millisecondsInterval)) {
            DoubleTimeObject before = null;
            DoubleTimeObject after = null;
            DoubleTimeObject middle = null;
            for (int indexList = tmpIndex; indexList < recordDataList.size(); indexList++) {
                DoubleTimeObject recordData = recordDataList.get(indexList);

                long span = recordData.data_time.getTime() - tmpDate.getTime();

                // 缓慢移动指针
                if (span < 0) {
                    tmpIndex = indexList;
                }

                if (span > -millisecondsInterval && span < 0) {
                    before = recordData;
                } else if (span == 0) {
                    middle = recordData;
                } else if (span > 0) {
                    if (span < millisecondsInterval) {
                        after = recordData;
                    }
                    break;
                }
            }

            DoubleTimeObject intervalData = null;
            if (!fjd_interpolate) {
                if ("left".equalsIgnoreCase(ConstantCE.Inst_fjd_use_direction)) {
                    if (middle != null) {
                        intervalData = new DoubleTimeObject(tmpDate, 0.0);
                        intervalData.data_value = middle.data_value;
                    } else {
                        if (before != null) {
                            intervalData = new DoubleTimeObject(tmpDate, 0.0);
                            intervalData.data_value = before.data_value;
                        }
                    }
                } else if ("right".equalsIgnoreCase(ConstantCE.Inst_fjd_use_direction)) {
                    if (middle != null) {
                        intervalData = new DoubleTimeObject(tmpDate, 0.0);
                        intervalData.data_value = middle.data_value;
                    } else {
                        if (after != null) {
                            intervalData = new DoubleTimeObject(tmpDate, 0.0);
                            intervalData.data_value = after.data_value;
                        }
                    }
                } else {
                    if (middle != null) {
                        intervalData = new DoubleTimeObject(tmpDate, 0.0);
                        intervalData.data_value = middle.data_value;
                    } else if (before != null && after != null) {
                        if (tmpDate.getTime() - before.data_time.getTime() <= after.data_time.getTime() - tmpDate.getTime()) {
                            intervalData = new DoubleTimeObject(tmpDate, 0.0);
                            intervalData.data_value = before.data_value;
                        } else {
                            intervalData = new DoubleTimeObject(tmpDate, 0.0);
                            intervalData.data_value = after.data_value;
                        }
                    } else {
                        if (before != null) {
                            intervalData = new DoubleTimeObject(tmpDate, 0.0);
                            intervalData.data_value = before.data_value;
                        } else if (after != null) {
                            intervalData = new DoubleTimeObject(tmpDate, 0.0);
                            intervalData.data_value = after.data_value;
                        }
                    }
                }
            } else {
                if (middle != null) {
                    intervalData = new DoubleTimeObject(tmpDate, 0.0);
                    intervalData.data_value = middle.data_value;
                } else if (before != null && after != null) {
                    intervalData = new DoubleTimeObject(tmpDate, 0.0);
                    intervalData.data_value = before.data_value + (after.data_value - before.data_value)
                            * (tmpDate.getTime() - before.data_time.getTime()) / (after.data_time.getTime() - before.data_time.getTime());
                } else {
                    if (before != null && (tmpDate.getTime() - before.data_time.getTime()) * 3 <= millisecondsInterval) {
                        intervalData = new DoubleTimeObject(tmpDate, 0.0);
                        intervalData.data_value = before.data_value;
                    } else if (after != null && (after.data_time.getTime() - tmpDate.getTime()) * 3 <= millisecondsInterval) {
                        intervalData = new DoubleTimeObject(tmpDate, 0.0);
                        intervalData.data_value = after.data_value;
                    }
                }
            }

            if (intervalData != null) {
                result.add(intervalData);
            }
        }

        return result;
    }
}
