package com.sagacloud.util.split;

import com.sagacloud.util.compute_engine.util.DoubleTimeObject;

import java.util.List;

@SuppressWarnings({"rawtypes"})
public class SplitAlgorithm {
    private static final SplitAlgorithm instance = new SplitAlgorithm();

    private SplitAlgorithm() {

    }

    public static SplitAlgorithm Instance() {
        return instance;
    }

    private boolean DataError(List<Double> estimateValue, double energy, List resultValue) {
        // 有空值直接平均
        boolean has_null = false;
        for (int i = 0; i < estimateValue.size(); i++) {
            Double estimate = estimateValue.get(i);
            if (estimate == null) {
                has_null = true;
                break;
            }
        }
        if (has_null) {
            // 全部平均
            double value = energy / estimateValue.size();
            for (int i = 0; i < estimateValue.size(); i++) {
                Object onedata = resultValue.get(i);
                ((DoubleTimeObject) onedata).data_value = value;
            }
            return true;
        }
        // 估计值总和等于0直接平均
        double totalEstimateValue = 0.0;
        for (int i = 0; i < estimateValue.size(); i++) {
            totalEstimateValue += estimateValue.get(i);
        }
        if (totalEstimateValue == 0.0) {
            // 全部平均
            double value = energy / estimateValue.size();
            for (int i = 0; i < estimateValue.size(); i++) {
                Object onedata = resultValue.get(i);
                ((DoubleTimeObject) onedata).data_value = value;
            }
            return true;
        }

        return false;
    }

    public void Solve(List<Double> estimateValue, List<Double> uncertainty, double energy, List resultValue) {
        if (DataError(estimateValue, energy, resultValue)) {
            return;
        }

        double Sj2 = 0.0;
        double Xj = 0.0;
        for (int i = 0; i < estimateValue.size(); i++) {
            Sj2 += uncertainty.get(i) * uncertainty.get(i);
            Xj += estimateValue.get(i);
        }
        // 如果Sj2很小则平均分配
        if (Sj2 < 0.000001) {
            this.SolveByRatio(estimateValue, energy, resultValue);
        } else {
            for (int i = 0; i < estimateValue.size(); i++) {
                Object onedata = resultValue.get(i);
                double value = estimateValue.get(i) + (energy - Xj) * (uncertainty.get(i) * uncertainty.get(i)) / Sj2;
                ((DoubleTimeObject) onedata).data_value = value;

            }

            boolean negativeNo = true;
            for (int i = 0; i < estimateValue.size(); i++) {
                Object onedata = resultValue.get(i);
                double datavalue;
                datavalue = ((DoubleTimeObject) onedata).data_value;
                if (datavalue < 0.0) {
                    negativeNo = false;
                    break;
                }
            }
            if (!negativeNo) {
                SolveByRatio(estimateValue, energy, resultValue);
            }
        }

        // 检查拆分结果是否与总和一致，如果不一致，则所有末端集均摊
        double resultTotal = 0.0;
        for (int i = 0; i < estimateValue.size(); i++) {
            Object onedata = resultValue.get(i);
            double datavalue;
            datavalue = ((DoubleTimeObject) onedata).data_value;
            resultTotal += datavalue;
        }
        // 差0.1以内就算了
        double chazhi = Math.abs(resultTotal - energy);
        if (chazhi > 0.000001) {
            double shareValue = (energy - resultTotal) / estimateValue.size();
            for (int i = 0; i < estimateValue.size(); i++) {
                Object onedata = resultValue.get(i);
                ((DoubleTimeObject) onedata).data_value = shareValue;
            }
        }
    }

    public void SolveByRatio(List<Double> estimateValue, double energy, List resultValue) {
        if (DataError(estimateValue, energy, resultValue)) {
            return;
        }

        // 按比例分配
        double totalEstimateValue = 0.0;
        for (int i = 0; i < estimateValue.size(); i++) {
            totalEstimateValue += estimateValue.get(i);
        }
        for (int i = 0; i < estimateValue.size(); i++) {
            double value = energy * estimateValue.get(i) / totalEstimateValue;
            Object onedata = resultValue.get(i);
            ((DoubleTimeObject) onedata).data_value = value;
        }
    }

}
